# Copyright (c) 2023, Riverbank Computing Limited
# All rights reserved.
#
# This copy of SIP is licensed for use under the terms of the SIP License
# Agreement.  See the file LICENSE for more details.
#
# This copy of SIP may also used under the terms of the GNU General Public
# License v2 or v3 as published by the Free Software Foundation which can be
# found in the files LICENSE-GPL2 and LICENSE-GPL3 included in this package.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ('AS IS'
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from ...specification import ValueType

from .argument import ArgumentFormatter
from .base_formatter import BaseFormatter
from .enum import EnumFormatter
from .variable import VariableFormatter


class ValueListFormatter(BaseFormatter):
    """ This creates various string representations of a list of values. """

    @property
    def cpp_expression(self):
        """ The C++ representation of the value list as an expression. """

        return self._expression()

    def py_expression(self, embedded=False, as_xml=False):
        """ The Python representation of the value list as an expression. """

        return self._expression(as_python=True, embedded=embedded,
                as_xml=as_xml)

    def as_rest_ref(self, as_xml=False):
        """ Return the Python representation of the value list as a reST
        reference.
        """

        value_list = self.object

        # The value must be a scoped name and we don't handle expressions.
        if len(value_list) != 1 or value_list[0].value_type is not ValueType.SCOPED:
            return None

        target = value_list[0].value

        # See if it is an attribute.
        for variable in self.spec.variables:
            if variable.fq_cpp_name == target:
                return VariableFormatter(self.spec, variable).as_rest_ref()

        # See if it is an enum member.
        target_scope = target.scope
        if target_scope is not None:
            target_scope.make_absolute()

        target_base_name = target.base_name

        for enum in self.spec.enums:
            # Look for the member name first before working out if it is the
            # correct enum.
            for member in enum.members:
                if member.cpp_name == target_base_name:
                    formatter = EnumFormatter(self.spec, enum)

                    if enum.is_scoped:
                        # It's a scoped enum so the fully qualified name of the
                        # enum must match the scope of the name.
                        if target_scope is not None and enum.fq_cpp_name == target_scope:
                            return formatter.member_as_rest_ref(member)
                    else:
                        # It's a traditional enum so the scope of the enum must
                        # match the scope of the name.
                        if (enum.scope is None and target_scope is None) or (enum.scope is not None and target_scope is not None and enum.scope.iface_file.fq_cpp_name == target_scope):
                            return formatter.member_as_rest_ref(member)

                    break

        return None

    def _expression(self, as_python=False, embedded=False, as_xml=False):
        """ The representation of the value list as an expression. """

        s = ''

        for value in self.object:
            if value.cast is not None and not as_python:
                s += '(' + value.cast.as_cpp + ')'

            if value.unary_operator is not None:
                s += value.unary_operator

            if value.value_type is ValueType.QCHAR:
                if value.value == '"' and embedded:
                    s += "'\\\"'"
                elif value.value.isprintable():
                    s += "'" + value.value + "'"
                else:
                    s += f"'\\{ord(value.value):03o}'"

            elif value.value_type is ValueType.STRING:
                quote = "\\\"" if embedded else "\""

                s += quote

                for ch in value.value:
                    escape = True

                    if ch in "\\\"":
                        pass
                    elif ch == '\n':
                        ch = 'n';
                    elif ch == '\r':
                        ch = 'r'
                    elif ch == '\t':
                        ch = 't'
                    else:
                        escape = False

                    if escape:
                        s += "\\"

                    s += ch

                s += quote

            elif value.value_type is ValueType.NUMERIC:
                s += str(value.value) if as_python else str(int(value.value))

            elif value.value_type is ValueType.REAL:
                s += format(value.value, 'g')

            elif value.value_type is ValueType.SCOPED:
                if as_python:
                    s += value.value.as_py
                else:
                    s += value.value.as_cpp

            elif value.value_type is ValueType.FCALL:
                arg_formatter = ArgumentFormatter(self.spec,
                        value.value.result)

                if as_python:
                    s += arg_formatter.as_py_type()
                else:
                    s += arg_formatter.cpp_type()

                args = [ValueListFormatter(self.spec, a)._expression(
                                as_python=as_python, embedded=embedded,
                                as_xml=as_xml)
                        for a in value.value.args]

                separator = ',' if as_xml else ', '
                s += '(' + separator.join(args) + ')'

            elif value.value_type is ValueType.EMPTY:
                s += '{}'

            if value.binary_operator is not None:
                s += value.binary_operator

        return s



# Copyright (c) 2022, Riverbank Computing Limited
# All rights reserved.
#
# This copy of SIP is licensed for use under the terms of the SIP License
# Agreement.  See the file LICENSE for more details.
#
# This copy of SIP may also used under the terms of the GNU General Public
# License v2 or v3 as published by the Free Software Foundation which can be
# found in the files LICENSE-GPL2 and LICENSE-GPL3 included in this package.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from abc import ABC, abstractmethod
import importlib
import importlib.util
import os

from .exceptions import UserException
from .pyproject import PyProject, PyProjectOptionException


class AbstractProject(ABC):
    """ This specifies the API of a project. """

    @classmethod
    def bootstrap(cls, tool, tool_description='', arguments=None):
        """ Return an AbstractProject instance fully configured for a
        particular command line tool.
        """

        # Get the contents of the pyproject.toml file.
        pyproject = PyProject()

        # Get the name of the project factory.
        project_factory_name = None

        sip_section_name = 'tool.sip'
        value_name = 'project-factory'

        sip_section = pyproject.get_section(sip_section_name)
        if sip_section is not None:
            f_name = sip_section.get(value_name)
            if f_name is not None:
                if not isinstance(f_name, str):
                    raise PyProjectOptionException(value_name,
                            "should be a 'str' and not '{0}'".format(
                                    type(f_name).__name__),
                            section_name=sip_section_name)

                project_factory_name = f_name

        # See if there is a corresponding callable.
        if project_factory_name is None:
            default_factory_py = 'project.py'

            if os.path.isfile(default_factory_py):
                project_factory = cls.import_callable(default_factory_py, cls)
            else:
                # The default project factory.
                from .project import Project as project_factory
        else:
            project_factory = cls.import_callable(project_factory_name, cls)

        project = project_factory()

        if not isinstance(project, cls):
            raise UserException(
                    "The project factory did not return an AbstractProject "
                    "object")

        # We set this as an attribute rather than change the API of the ctor or
        # setup().
        project.arguments = arguments

        # Complete the configuration of the project.
        project.setup(pyproject, tool, tool_description)

        return project

    @abstractmethod
    def build(self):
        """ Build the project in-situ. """

    @abstractmethod
    def build_sdist(self, sdist_directory):
        """ Build an sdist for the project and return the name of the sdist
        file.
        """

    @abstractmethod
    def build_wheel(self, wheel_directory):
        """ Build a wheel for the project and return the name of the wheel
        file.
        """

    @staticmethod
    def import_callable(name, base_type):
        """ Import a callable from either a .py file or specified as
        module[:object].
        """

        # See if the name refers to a .py file.
        if name.endswith('.py'):
            name = name.replace('/', os.sep)

            module_name = name[:-3]
            object_name = None

            # Try and import the .py file.
            spec = importlib.util.spec_from_file_location(module_name, name)
            module = importlib.util.module_from_spec(spec)

            try:
                spec.loader.exec_module(module)
            except Exception as e:
                raise UserException("Unable to import '{0}'".format(name),
                        detail=str(e))
        else:
            # Extract the module and any object name.
            parts = name.split(':')
            if len(parts) > 2:
                raise UserException(
                        "The callable '{0}' must be specified as "
                        "'module[:name]'".format(name))

            module_name = parts[0]
            object_name = parts[1] if len(parts) == 2 else None

            # Try and import the module.
            try:
                module = importlib.import_module(module_name)
            except ImportError as e:
                raise UserException(
                        "Unable to import '{0}'".format(module_name),
                        detail=str(e))

        # Get the callable object from the module.
        if object_name is None:
            # Look for a class that is a sub-class of the base type.
            for obj in module.__dict__.values():
                if isinstance(obj, type):
                    if issubclass(obj, base_type):
                        # Make sure the type is defined in the module and not
                        # imported by it.
                        if obj.__module__ == module_name:
                            break
            else:
                raise UserException(
                        "'{0}' does not define a {1} sub-class".format(name,
                                base_type.__name__))
        else:
            # We have the name of the callable so just get it.
            obj = getattr(module, object_name)
            if obj is None:
                raise UserException(
                        "'{0}' module has no callable '{1}'".format(
                                module_name, object_name))

        return obj

    @abstractmethod
    def install(self):
        """ Install the project. """

    @abstractmethod
    def setup(self, pyproject, tool, tool_description):
        """ Complete the configuration of the project. """
